#!/usr/bin/env python3 -*- coding: utf-8

import serial
import os
import sys
from time import sleep
from datetime import datetime
from subprocess import PIPE, run
#from threading import Event, Thread

#__________________________________________________
#
# SETUP:


serial_port = "/dev/ttyACM0"
fifo_path = "/tmp/pyserial.fifo"
fifo = ''

admin_dir = os.path.splitext('/home/pi/ADMIN')[0]
admin_states = os.path.splitext(admin_dir + '/.sys/.states')[0]

cafca_dir = os.path.splitext('/home/pi/CAFCA')[0]
cafca_states = os.path.splitext(cafca_dir + '/data/.states')[0]

log_dir = os.path.splitext('/home/pi/ADMIN/.log')[0]
log_file = (str(os.path.splitext(log_dir + '/serialpy')[0])+'.log')
writeable = True
hide = " >/dev/null 2>&1"

#--------------------------------------------------


try:
    os.mkfifo(fifo_path)
    print("fifo succesfully created!")
    pass

except FileExistsError:
    print("fifo already exists!")
    os.remove(fifo_path)
    sleep(1)
    os.mkfifo(fifo_path)
    sleep(1)
    pass
except OSError as e:
    print(f"fifo creation failed: {e}")
    #pass

try:
    print("opening fifo")
    fifo = open(fifo_path, "r")
    sleep(1)
    os.system('printf "VFD 1 TEST" > /tmp/pyserial.fifo')
    pass

except Exception as e:
    print("error opening fifo:",e)
    #pass

def setup():
    print("SETUP")
    try:
        if not os.path.exists(log_file):
            os.system('sudo touch ' + log_file)
            sleep(1)

        os.system('sudo chmod 775 ' + log_file)
        sleep(1)

    except Exception as e:
        print("error writing logfile:",e)
        writeable = False

    os.system('echo \x22\x23 SERIAL LOG\n\x23 ' + str(f'{datetime.today():%h %e %Y}') + '\n\x22 | sudo tee ' + log_file + hide)

    write_log("_SETUP","DONE")
#__________________________________________________
#
# FUNCTIONS:

def loop():

    if os.path.exists(serial_port):
        print("Found serial port",serial_port,"!")
        sleep(1)

        with serial.Serial('/dev/ttyACM0', 9600, timeout=1) as arduino:
            try:
                sleep(0.1)

                if arduino.isOpen():
                    arduino.reset_input_buffer()
                    send_fifo('BOOTED 1')
                    send_fifo('VFD 1 TEST')

                    try:
                        if arduino.inWaiting()==0:
                            msg = read_fifo()

                            if len(msg)>0:
                                arduino.write(msg.encode())

                        elif arduino.inWaiting()>0:
                            line = arduino.readline().decode('utf-8').strip()

                            write_log("TEENSY",line)
                            nom = line.split()[0]

                            arduino.flushInput()
                            checkInbox(line)

                    except serial.SerialException as se:
                        print("Serial port error:",se)
                        #break

            except Exception as e:
                print(f"Error opening Arduino:",e)

            except KeyboardInterrupt:
                print("KeyboardInterrupt has been caught.")
                arduino.close()
                fifo.close()
                os.remove(fifo_path)
                #os._exit(1)
    else:
        print(f"Waiting for {serial_port}...")
        sleep(5)


#--------------------------------------------------

def send_fifo(msg):
    os.system('printf \x22'+ msg + '\x22 > /tmp/pyserial.fifo')

def checkInbox(input):
    caller = input.split()[0]

    if caller == 'RPI':
        pos = input.find('RPI') + 4
        msg = input[pos:]

        if msg == "WAKEUP":
            os.system("sed -i '/ASLEEP/d' " + admin_states)
            sleep(0.5)
            os.system("echo 'BOOL ASLEEP=0' | sudo tee -a " + admin_states + hide)

        if msg == "GAME_LOADED 1":
            os.system("sed -i '/GAME_LOADED/d' " + cafca_states)
            sleep(0.5)
            os.system("echo 'BOOL GAME_LOADED=1' | sudo tee -a " + cafca_states + hide)

        if msg == "GAME_STOPPED 1":
            os.system("sed -i '/GAME_LOADED/d' " + cafca_states)
            sleep(0.5)
            os.system("echo 'BOOL GAME_LOADED=0' | sudo tee -a " + cafca_states + hide)
    else:
        line = input;
        if line == "RESET":
            send_fifo('VFD 1 RPI SAYS HELLO')

    #if caller == 'AVR':
    #    pos = input.find('AVR') + 4
    #    line = input[pos:]


def read_fifo():
    output = ''
    line = fifo.readline()

    if line:
        output = line.replace('\n', '')
        output = output + ";"
        write_log('_FIFO>',line)
    return str(output)

#--------------------------------------------------

def write_log(src = ' ', msg = ' '):
    timestr = f'{datetime.today():%T}'

    if writeable:
        os.system("echo \x22[" + timestr + "] " + src + ": \x27" + msg + "\x27\x22 | sudo tee -a " + log_file) #+ hide)

#__________________________________________________

setup()
#loop()
while True:
    loop()

fifo.close()
arduino.close()
sys.exit()
